/** 
 * SALSA 1.1
 * June 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.awt.*;
import java.awt.event.*;

public class MemoryCanvas extends CritterCanvas {
    
  private static final int
    NAMES_WIDTH = 13,
    LEGEND_HEIGHT = 36,
    WIDTH = 300,
    HEIGHT = 264;
    
  private static final double
    OUTPUT_GAIN = 5.0;
    
  protected int
    legendY0,
    legendY1,
    legendY2,
    inputWidth,
    outputWidth;
    
  protected int[]
    sensorStarts,
    effectorStarts;
  
  protected Memory[] memories;
  
  protected Sensor[][] sensors;
  
  protected Effector[][] effectors;

  MemoryCanvas (UI ui, World world) {
      super(ui, world);
  } // constructor MemoryCanvas
  
  protected void init() {
      setMemories();
      sensors = getSensors();
      effectors = getEffectors();
  }
  
  /**
  * Has to precede setDimensions().
  */
  private void setMemories() {
      this.memories = getMemories();
      for (int c = 0; c < nCritters; c++)
	  memories[c].setCanvas(this);
  } // method setMemories()
  
  /**
  * First figures the number of cells for each Critter.
  * Then figures the Canvas width based on the Critter with the most cells.
  * Then goes back and figures the cellWidths for each Critter.   
  */
  protected void setDimensions() {
      Memory memory;
      int c;
      int maxNCells = 0;
      int maxCellWidth;
      
      sensorStarts = new int[Params.NSENSORS];
      
      effectorStarts = new int[Params.NEFFECTORS];
      
      namesWidth = NAMES_WIDTH;
      
      inputWidth = Species.getMax(Params.VISUAL_RESOLUTION)
      + Species.getMax(Params.TOUCH_RANGE)
      + Species.getMax(Params.SONAR_RESOLUTION);
      
      outputWidth = Species.getMax(Params.NREST) + Species.getMax(Params.NMOVE)
      + Species.getMax(Params.NTURN) + Species.getMax(Params.NEAT)
      + Species.getMax(Params.NHIT);
      
      cellHeight = HEIGHT / nCritters;
      mainHeight = cellHeight * nCritters;
      height = mainHeight + LEGEND_HEIGHT + 1;
      
      legendY0 = mainHeight + 12;
      legendY1 = legendY0 + 11;
      legendY2 = legendY1 + 11;
      
      cellWidth = WIDTH / (inputWidth + outputWidth);
      mainWidth = (inputWidth + outputWidth) * cellWidth;
      
      width = mainWidth + namesWidth;
      
      sensorStarts[Params.TOUCH] = 0;
      sensorStarts[Params.VISION] = Species.getMax(Params.TOUCH_RANGE) * cellWidth;
      sensorStarts[Params.SONAR] = sensorStarts[Params.VISION]
      + Species.getMax(Params.VISUAL_RESOLUTION) * cellWidth;
      
      effectorStarts[Params.REST] = inputWidth * cellWidth;
      effectorStarts[Params.EAT] = effectorStarts[Params.REST] +
      Species.getMax(Params.NREST) * cellWidth;
      effectorStarts[Params.MOVE] = effectorStarts[Params.EAT] +
      Species.getMax(Params.NEAT) * cellWidth;
      effectorStarts[Params.HIT] = effectorStarts[Params.MOVE] +
      Species.getMax(Params.NMOVE) * cellWidth;
      effectorStarts[Params.TURN] = effectorStarts[Params.HIT] +
      Species.getMax(Params.NHIT) * cellWidth;
      
  } // setDimensions()

  protected void drawLines(Graphics g) {
    g.setColor(BORDER_COLOR2);
    g.drawLine(sensorStarts[Params.VISION], 0, sensorStarts[Params.VISION], mainHeight);
    g.drawLine(sensorStarts[Params.SONAR], 0, sensorStarts[Params.SONAR], mainHeight);

    g.drawLine(effectorStarts[Params.TURN], 0, effectorStarts[Params.TURN], mainHeight);
    g.drawLine(effectorStarts[Params.MOVE], 0, effectorStarts[Params.MOVE], mainHeight);
    g.drawLine(effectorStarts[Params.EAT], 0, effectorStarts[Params.EAT], mainHeight);
    g.drawLine(effectorStarts[Params.HIT], 0, effectorStarts[Params.HIT], mainHeight);
    
    g.setColor(BORDER_COLOR);
    g.drawLine(0, 0, width, 0);
    g.drawLine(0, 0, 0, height - 1);
    g.drawLine(0, height - 1, width - 1, height - 1);
    g.drawLine(width - 1, 0, width - 1, height - 1);
    g.drawLine(inputWidth * cellWidth, 0, inputWidth * cellWidth, height - 1);

    g.drawLine(mainWidth, 0, mainWidth, height - 1);
  } // method initImage()
  
  protected void writeLegend(Graphics g) {
      int i;
      int y;
      g.setFont(FONT);
      for (i = 0; i < Params.NSENSORS; i++) {
	if ((i % 3) == 0)
	    y = legendY0;
	else if ((i % 3) == 1)
	    y = legendY1;
	else
	    y = legendY2;
        g.drawString(Params.SENSOR_LABELS[i],
		     sensorStarts[i] + 1,
		     y);
      }
      for (i = 0; i < Params.NEFFECTORS; i++) {
	if ((i % 3) == 0)
	    y = legendY0;
	else if ((i % 3) == 1)
	    y = legendY1;
	else
	    y = legendY2;
        g.drawString(Params.EFFECTOR_LABELS[i],
		     effectorStarts[i] + 1,
		     y);
      }

  } // writeLegend(Graphics)
  
  public void paintCritter(Graphics g, int critIndex) {
      if (!isDead(critIndex)) {
	  int y = cellHeight * critIndex;
	  int s, o, e, i;
	  Memory memory = memories[critIndex];
	  
	  if (memory.isDead()) {
	      dead[critIndex] = true;
	      g.setColor(DEAD_COLOR);
	      g.fillRect(1, y + 1, width - 2, cellHeight - 1);
	      
	  } else {
	      Sensor[] critSensors = sensors[critIndex];
	      Effector[] critEffectors = effectors[critIndex];
	      Sensor sensor;
	      Effector effector;
	      
	      for (s = 0; s < critSensors.length; s++) {
		  sensor = critSensors[s];
		  if (sensor != null)
		      sensor.paint(g, sensorStarts[s], y, cellWidth, cellHeight);
	      }
	      
	      for (e = 0, o = 0; e < critEffectors.length; e++) {
		  effector = critEffectors[e];
		  if (effector != null)
		      for (i = 0; i < effector.getNActions(); i++, o++) {
		      paintOutput(g, effectorStarts[e] + i * cellWidth,
				  y, cellWidth, memory.getOutput(o));
		  }
	      }
	  }
      }
  } // paintCritter(Graphics, int)
  
  private void paintHidden(Graphics g, int x, int y, int cellWidth, double value) {
      g.setColor(Utils.doubleToBlue(value));
      g.fillRect (x, y + 1, cellWidth, cellHeight - 1);
  }
  
  private void paintOutput(Graphics g, int x, int y, int cellWidth, double value) {
      g.setColor(Utils.doubleToGreenRed(value, OUTPUT_GAIN));
      g.fillRect (x + 1, y + 1, cellWidth - 1, cellHeight - 1);
  }

} // class MemoryCanvas
