/** 
 * SALSA 1.1
 * June 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.io.*;
import java.awt.Graphics;
import java.awt.Color;

/**
* Subclass of the abstract class Sensor which implements a primitive
* sonar system with a limited range and uses competitive learning
* to reduce the number of states to something manageable.
*/
public class Sonar extends Sensor {
    
  private static final Color
    RAY_COLOR = Color.orange,
    COLOR = Color.darkGray;

  private static final Color[]
    COLORS = {Cell.LIGHTEST_GRAY, Color.lightGray,
              Color.gray, Color.darkGray, Color.black};
    
  public static final double
    // Proportion of the width of the world that is the maximum
    // distance for Sonar.
    MAX_DISTANCE_PROP = 1.0;
    
  // Canvases in windows for displaying retina, competitive learning
  // network activations and weights.  Public so that paint methods
  // in canvases can be called here.

//  public SonarCanvas sonarCanvas = null;
    
  public int
    // Index of the Critter; needed to get the right row in
    // SonarCanvas.
    index = 0;

//  public SonarCanvas sonarCanvas = null;

  private double[] distances;
  
  private double angle;
  
  private int
    nDirections,
    distinctions;
    
 /**
  * This constructor sets the UI and the Critter, initializes
  * the retina and creates the frames (and canvases) in which
  * the retina, CL activations, and CL weights are displayed.
  */
  Sonar (UI ui, int species) {
    super (ui, species);
    nDirections = getSpecies().get(Params.SONAR_RESOLUTION);
//    nStates = getSpecies().get(Params.NSONAR_DISTINCTIONS);
    nStates = Params.NSONAR;
    length = nDirections * nStates;
    distances = new double[nDirections];
    angle = getSpecies().get(Params.SONAR_ANGLE) * Math.PI / 180.0;
  } // constructor Sonar
  
  // ACCESS METHODS.
  
  public int getUnitsPerPosition() {
      return nStates;
  } // getUnitsPerPosition()

  /**
   * Returns a color representing the reading for a given direction
   * (ray, position).
   */
  protected Color distToColor(int dist) {
      if (dist >= COLORS.length)
	  return COLORS[COLORS.length - 1];
      else
	  return COLORS[dist];
  } // dist2Color(int)
  
  protected Color getColor(int dist) {
      if ((dist + 1) >= COLORS.length)
	  return COLORS[COLORS.length -1];
      else
	  return COLORS[dist + 1];
  }

  private Graphics getWorldGraphics() {
      return getWorldCanvas().getGraphics();
  } // method getWorldGraphics()
  
  private WorldCanvas getWorldCanvas() {
      return ui.world.canvas;
  } // method getWorldCanvas()
  
  protected Color getDisplayColor(int i) {
      return distToColor(Utils.roundDoubleToInt(distances[i], nStates));
  } // method getDisplayColor(int)
  
  protected void setDisplayLength() {
      displayLength = getSpecies().get(Params.SONAR_RESOLUTION);
//      displayLength = Params.getCrit(species, Params.SONAR_RESOLUTION);
  } // method setDisplayLength()
  
  private int[] getPixCoords(int x, int y, double rad, double ang) {
      int[] coords = new int[2];
      
      coords[0] = x + (int)(rad * Math.sin(ang));
      coords[1] = y - (int)(rad * Math.cos(ang));
      
      if (coords[0] < 0)
	  coords[0] += getWorldCanvas().getWidth();
      else if (coords[0] > getWorldCanvas().getWidth())
	  coords[0] -= getWorldCanvas().getWidth();
	  
      if (coords[1] < 0)
	  coords[1] += getWorldCanvas().getHeight();
      else if (coords[1] > getWorldCanvas().getHeight())
	  coords[1] -= getWorldCanvas().getHeight();
      
      return coords;
  } // method getPixCoords

  private void initDistances() {
      for (int p = 0; p < nDirections; p++)
        distances[p] = -1.0;
  } // method initDistances()
  
  private void getRayArcs(Cell cell, int heading) {
      int[] coords;
      int x = cell.getCenterX(),
          y = cell.getCenterY(),
	  r,   // Ray index
	  s,   // Sample index
	  D,   // Distance (3rds of Cell radius)
	  rad = 0; // Actual distance for the arc
	  
      double angle0,
	     angIncr;
      boolean anyEmpty = true;
      
      if (nDirections == 1) {
	  angle0 = Cell.heading2Angle(heading);
	  angIncr = 0.0;
      } else {
       angle0 = Cell.heading2Angle(heading)  // - angle / 2.0;
         - (angle / nDirections) * ((nDirections - 1) / 2);
       angIncr = // angle / (nDirections - 1);
         angle / nDirections;
      }
	     
      initDistances();
      
      // Start at a distance of 2 because the closest arc overlaps with
      // the Critter's own Cell.
      D = 2;

      while (anyEmpty && rad < Params.maxCellDistance) {
	  rad = (D * World.getCellRadius()) / 2;
	  anyEmpty = false;
	  for (r = 0; r < nDirections; r++) {
              if (distances[r] == -1.0) {
		// Not already obscured by something closer.
  	        coords = getPixCoords(x, y, rad, angle0 + r * angIncr);
	        if (coords != null && Pixel.getThing(coords, getWorldCanvas()) != null) {
		    // The point is not outside the World
		    distances[r] = scaleDistance(rad);
//		    ui.write("  distance to obj: " + distances[r] + "(" + rad + ")");
	        }
   	        else if (ui.showRays() && coords != null)
   		    Pixel.paint(coords, getWorldGraphics(), RAY_COLOR);
              }
              
	      if (distances[r] == -1.0)
		  anyEmpty = true;
	  }
	  D++;
      }
  } // method getRayArcs(Cell, int)
  
  /**
  * Converts an actual pixel-to-pixel distance to an exaggerated proportion,
  * 0 for maximum, 1 for minimum distance.   
  */
  private static double scaleDistance(int dist) {
      return
        Utils.reverseExaggerate
          (Utils.getRangeProportion
             (Params.minCellDistance,
	      Params.maxCellDistance,
	      dist));
  } // method scaleDistance(int)
  
  private int[] distancesToPercept() {
      int[] sensed = new int[length];
      int dir, pStart;
      double distance;
      
      for (dir = 0, pStart = 0;
	   dir < nDirections;
	   dir++, pStart += nStates) {
	  Utils.double2Thermometer(distances[dir], sensed,
				   pStart, nStates);
      }
      return sensed;
  } // method distancesToPercept()
  
  public int[] sense(Cell cell, int direction) {
      int[] sensed = new int[length];
      
      getRayArcs(cell, direction);
      sensed = distancesToPercept();
      
//      Utils.writeIntArray(sensed, ui);
      
      return sensed;
  } // method sense(Cell, int)
  
  protected void paint(Graphics g, int i, int x, int y, int w, int h) {
//      int barHeight =
//        (int)(Utils.roundDouble(distances[i], nStates) * (h - 1));
//      g.fillRect(x + 1, y + h - barHeight, w - 1, barHeight);
      g.fillRect(x + 1, y + 1, w - 1, h - 1);
  } // method point(Graphics, int, int, int, int, int)

  protected void clear(Graphics g, int x, int y, int w, int h) {
//    g.setColor(MemoryCanvas.BACKGROUND_COLOR);
//    g.fillRect(x + 1, y + 1, w - 1, h - 1);
  }
} // class Sonar
