/** 
 * SALSA 1.1
 * June 2000
 * Mike Gasser (gasser@indiana.edu)
 */

import java.io.*;

public class Memory {

  private static final double
    GAIN = 1.0,
    INIT_WEIGHT_RANGE = 1.0,
    INIT_OUTPUT = 0.5;
    
  private static final double
    REINF_MULT = .02;
    
  private static final double
    LEARNING_RATE_DIV = 100.0,
    EXPLOITATION_RATE_DIV = 1000.0,
    DISCOUNT_RATE_DIV = 10.0;

  private int
    lastAction = 0;
    
  private int[]
    lastPercept;
    
  private UI ui;
  
  private Sensor[] sensors;
  
  private Effector[] effectors;
  
  private MemoryCanvas canvas = null;
  
  private WeightCanvas weightCanvas = null;
    
  private String[] stateNames;
    
  private int
    // Excludes the bias unit
    nInputs,
    nHidden,
    nOutputs,
    nWeights = 0;
    
  private int[][]
    outputs2Actions;

  private double[]
    inputs,
    hidden,
    outputs,
    hiddenDeltas,
    outputDeltas;
    
  private double[][]
    in2OutWeights = null,
    hid2OutWeights = null,
    in2HidWeights = null;

  public boolean
    fresh = true,
    dead = false;
    
  private double
    learningRate,
    discountRate,
    exploitationRate;
    
  private int
    species;

  Memory(UI ui) {
    this.ui = ui;
  } // constructor Memory(UI)
  
  Memory(UI ui, Effector[] effectors, int nInputs, int nHidden, int species) {
    this.ui = ui;
    this.nInputs = nInputs;
    this.nHidden = nHidden;
    setSpecies(species);
    setParams();
    // Also sets nOutputs
    setEffectors(effectors);
    init();
  } // constructor Memory
  
  // GETTERS
  
  public int getNWeights() {
      return nWeights;
  } // method getNWeights()
  
  private Species getSpecies() {
      return Params.getSpecies(species);
  } // getSpecies()

  /**
  * Get the Q value for the action with index i.
  */
  private double getQ(int i) {
      return outputs[i];
  } // method getQ(int)
  
  private double getMaxQ(int[] percept) {
      ui.write("Selecting highest Q for percept", 1);
      
      activate(percept);
      
      return getMaxQ();
  } // method getMaxQ(int[])
  
  private double getMaxQ() {
      double maxQ = -100.0, Q;
      for (int i = 0; i < nOutputs; i++) {
          Q = getQ(i);
          if (Q > maxQ)
	      maxQ = Q;
      }
      return maxQ;
  } // method getMaxQ()
  
  private double getNewQ(int[] percept, double reinforcement) {
      return (
      // Reinforcement
      reinforcement +
      // Discounted best value of this state
      discountRate * getMaxQ(percept));
  } // method getNewQ(int[], int)
  
  public boolean isDead() {
    return dead;
  } // method isDead()

  public void setSpecies(int s) {
      species = s;
  }
  
  public void setParams() {
      learningRate =
        getSpecies().get(Params.LEARNING_RATE) / LEARNING_RATE_DIV;
      discountRate =
        getSpecies().get(Params.DISCOUNT_RATE) / DISCOUNT_RATE_DIV;
      exploitationRate =
        getSpecies().get(Params.EXPLOITATION_RATE) / EXPLOITATION_RATE_DIV;
//      ui.write("Exploitation rate: " + exploitationRate);
  } // method setParams()

  /**
  * Sets the effectors, nOutputs, and the mapping from output units
  * to actions (outputs2Actions).   
  */
  private void setEffectors(Effector[] effectors) {
      int e, i, o;
      Effector effector;
      
      this.effectors = effectors;
      nOutputs = 0;
      
      for (e = 0; e < effectors.length; e++) {
          effector = getEffector(e);
          if (effector != null)
  	    nOutputs += getEffector(e).getNActions();
      }
      
      outputs2Actions = new int[nOutputs][2];
      for (e = 0, o = 0; e < effectors.length; e++) {
          effector = getEffector(e);
	  if (effector != null)
            for (i = 0; i < effector.getNActions(); i++, o++) {
	        outputs2Actions[o][0] = e;  // Effector index
	        outputs2Actions[o][1] = i;  // Action index
            }
      }
  } // method setEffectors(Effector[])
  
  public Memory clone(Critter c) {
      Memory m = new Memory(ui, c.getEffectors(), nInputs, nHidden, species);
//      ui.write("Cloning memory for Critter");
      copyWeights(m);
      return m;
  } // clone(Critter)
  
  private void copyWeights(Memory m) {
      int i, h, o;
      if (nHidden > 0) {
	 for (h = 0; h < nHidden; h++)
	     for (i = 0; i < nInputs + 1; i++)
		 m.setIn2Hid(h, i, getIn2Hid(h, i));
      }
      for (o = 0; o < nOutputs; o++) {
        for (i = 0; i < nInputs + 1; i++)
	    m.setIn2Out(o, i, getIn2Out(o, i));
	if (nHidden > 0) {
	 for (h = 0; h < nHidden; h++)
	     m.setHid2Out(o, h, getHid2Out(o, h));
	}
      }
  } // copyWeights(Memory)

  public void init() {
    int i, h, o;
    inputs = new double[nInputs + 1];
    outputs = new double[nOutputs];
    outputDeltas = new double[nOutputs];
    // Output units are not connected to input units except for the bias unit
    // unless there are no hidden units.
    in2OutWeights = new double[nOutputs][nInputs + 1];
    if (nHidden > 0) {
      hidden = new double[nHidden];
      hiddenDeltas = new double[nHidden];
      hid2OutWeights = new double[nOutputs][nHidden];
      in2HidWeights = new double[nHidden][nInputs + 1];
    }
    
    for (i = 0; i < nInputs; i++)
      inputs[i] = Utils.booleanToDouble(Utils.getRandomBoolean());

    // Bias activation
    inputs[nInputs] = 1.0;

    for (h = 0; h < nHidden; h++) {
      hidden[h] = INIT_OUTPUT;
      hiddenDeltas[h] = 0.0;
      for (i = 0; i < nInputs + 1; i++) {
        in2HidWeights[h][i] = makeInitWeight();
	nWeights++;
      }
    }

    for (o = 0; o < nOutputs; o++) {
      outputs[o] = INIT_OUTPUT;
      outputDeltas[o] = 0.0;
      for (i = 0; i < nInputs + 1; i++) {
	 in2OutWeights[o][i] = makeInitWeight();
	 nWeights++;
      }
      if (nHidden > 0) {
        for (h = 0; h < nHidden; h++) {
          hid2OutWeights[o][h] = makeInitWeight();
	  nWeights++;
        }
      }
    }

    lastPercept = new int[nInputs];
    for (i = 0; i < nInputs; i++)
      lastPercept[i] = 0;
  } // init()
  
  public void reinit() {
    int i, h, o;
    
//    ui.write("nInputs: " + nInputs + ", nOutputs: " + nOutputs);
    
    for (i = 0; i < nInputs; i++)
      inputs[i] = Utils.booleanToDouble(Utils.getRandomBoolean());

    // Bias activation
    inputs[nInputs] = 1.0;

    for (h = 0; h < nHidden; h++) {
      hidden[h] = INIT_OUTPUT;
      hiddenDeltas[h] = 0.0;
      for (i = 0; i < nInputs + 1; i++) {
        in2HidWeights[h][i] = makeInitWeight();
      }
    }

    for (o = 0; o < nOutputs; o++) {
      outputs[o] = INIT_OUTPUT;
      outputDeltas[o] = 0.0;
      for (i = 0; i < nInputs + 1; i++) {
	 in2OutWeights[o][i] = makeInitWeight();
//	 ui.write("in2OutWeights" + o + i + ": " +
//		  in2OutWeights[o][i]);
      }
      if (nHidden > 0) {
        for (h = 0; h < nHidden; h++) {
          hid2OutWeights[o][h] = makeInitWeight();
        }
      }
    }

    for (i = 0; i < nInputs; i++)
      lastPercept[i] = 0;
  
    dead = false;
  } // reinit()
  
  private double makeInitWeight() {
      return 0.0;
  } // makeInitWeight()
  
  private Sensor getSensor() {
      return sensors[0];
  }
  
  private Effector getEffector(int i) {
      return effectors[i];
  }

  // ACCESS METHODS FOR NETWORKS
  
  public int getNInputs() {
      return nInputs;
  }
  
  public int getNHidden() {
      return nHidden;
  }
  
  public int getNOutputs() {
      return nOutputs;
  }
  
  public int getNIn2Hid() {
      if (in2HidWeights != null)
        return in2HidWeights.length * in2HidWeights[0].length;
      else
	return 0;
  }
  
  public int getNIn2Out() {
      if (in2OutWeights != null)
        return in2OutWeights.length * in2OutWeights[0].length;
      else
	return 0;
  }
  
  public int getNHid2Out() {
      if (hid2OutWeights != null)
        return hid2OutWeights.length * hid2OutWeights[0].length;
      else
	return 0;
  }
  
  public double getInput(int index) {
      return inputs[index];
  }
  
  public double getHidden(int index) {
      return hidden[index];
  }
  
  public double getOutput(int index) {
      return outputs[index];
  }
  
  public double getIn2Hid(int hid, int in) {
      return in2HidWeights[hid][in];
  }
  
  public double getIn2Out(int out, int in) {
      return in2OutWeights[out][in];
  }
  
  public double[] getIn2Out(int out) {
      return in2OutWeights[out];
  }
  
  public double getHid2Out(int out, int hid) {
      return hid2OutWeights[out][hid];
  }
  
  public void setIn2Hid(int h, int i, double value) {
      in2HidWeights[h][i] = value;
  }
  
  public void setIn2Out(int o, int i, double value) {
      in2OutWeights[o][i] = value;
  }
  
  public void setHid2Out(int o, int h, double value) {
      hid2OutWeights[o][h] = value;
  }
  
  public void setCanvas(MemoryCanvas canvas) {
      this.canvas = canvas;
  }

  public void setWeightCanvas(WeightCanvas canvas) {
      this.weightCanvas = canvas;
  }

  private double makeRandomWeight() {
     return (Math.random() * INIT_WEIGHT_RANGE - .5 * INIT_WEIGHT_RANGE);
  } // method makeRandomWeight()
  
  public void show() {
      canvas.repaint();
  }
  
  private void run(double[] input, double[] target) {
      clamp(input);
      if (nHidden > 0)
        activateHidden();
      activateOutput();
      figureError(target);
      if (nHidden > 0)
        figureError();
      updateOutputWeights();
      if (nHidden > 0)
        updateHiddenWeights();
  } // method run(double[], double[])
  
  /**
   * Like the above except input is int[], and
   * target is just for one output position (targetI)
   */
  private void run(int[] input, double target, int targetI) {
      activate(input);
      // Now do backprop
      figureError(targetI, target);
      if (nHidden > 0)
        figureError();
      updateOutputWeights();
      if (nHidden > 0)
        updateHiddenWeights();
  } // method run(int[], double, int)

  public void activate() {
      if (nHidden > 0)
        activateHidden();
      activateOutput();
  } // method activate()

  public void activate(int[] input) {
    setPercept(input);
    activate();
  }

  private void activateHidden() {
      int i;
      for (i = 0; i < nHidden; i++)
	  hidden[i] = activateHidden(getInput(i, inputs, in2HidWeights), GAIN);
  } // method activateHidden()
  
  private void activateOutput() {
      int i;
      double in;
      for (i = 0; i < nOutputs; i++) {
	if (nHidden > 0) {
            in = getInput(i, hidden, hid2OutWeights);
            in += in2OutWeights[i][0];
	} else
	    in = getInput(i, inputs, in2OutWeights);

	outputs[i] = activateOutput(in, GAIN);
      }
  } // method activateOutput()
  
  /**
  * Linear activation function for output units.
  */
  private double activateOutput(double i, double gain) {
    return i;
  } // method activateOutput(double, double)
  
  private double activateHidden(double i, double gain) {
      return Utils.squash(i, gain);
  } // method activateHidden(double, double)
  
  private double activateFunction(double i, double gain) {
     return Utils.squash(i, gain);
  } // method activationFunction(double, double)

  private double getInput(int i, double[] activations, double[][] weights) {
    return (Utils.dotProduct (activations, weights[i]));
  } // method getInput(int, double[], double[][])

  /**
   * Length of input should be nInputs, one less than
   * the length of inputs, because of the extra bias unit.   
   */
  private void clamp(double[] input) {
      int i = 0;
      while (i < nInputs && i < input.length) {
	  inputs[i] = input[i];
	  i++;
      }
  } // method clamp(int[])

  /**
  * Error for output layer.
  */
  private double figureError(double[] target) {
    int i;
    double e = 0.0;
    for (i = 0; i < nOutputs; i++) {
      e += figureUnitError(i, target[i]);
    }
    return e;
  }
  
  /**
  * Figure error for only a single output unit, setting other output deltas to 0.
  */
  private double figureError(int index, double target) {
      int i;
      double e = 0.0;
      for (i = 0; i < nOutputs; i++) {
          if (i == index)
              e = figureUnitError(i, target);
          else
	      outputDeltas[i] = 0;
      }
//      ui.write("Target: " + target + ", error: " + e);
      return e;
  }

 /**
  * Error for output unit
  */
  private double figureUnitError(int i, double target) {
    double e = target - outputs[i];

    outputDeltas[i] = e * figureOutputActivationSlope(i, outputs);

    ui.write ("Output unit " + i + "; error: " + e
		+ ", delta: " + outputDeltas[i], 1);

    return (e * e);
  }
  
  /**
  * Error for hidden layer.
  */
  private void figureError() {
    int i;
    for (i = 0; i < nHidden; i++) {
      figureUnitError(i);
    }
  }

  public void figureUnitError(int i) {
    hiddenDeltas[i] =  
      figureDeltaInput(i) * figureHiddenActivationSlope(i, hidden);
    ui.write("Hidden unit " + i + "; delta: " + hiddenDeltas[i], 1);
  }

  private double figureDeltaInput(int i) {
    double sum = 0.0;
    int o;

    for (o = 0; o < nOutputs; o++)
      sum += outputDeltas[o] * hid2OutWeights[o][i];

    return sum;
  }

  private double figureOutputActivationSlope(int i, double[] activations) {
    return 1.0;
  } // figureOutputActivationSlope(int, double[])
  
  private double figureHiddenActivationSlope(int i, double[] activations) {
    // for sigmoid squashing
    double act = activations[i];
    return (act * (1.0 - act));
  } // figureHiddenActivationSlope(int, double[])
  
  private void updateWeight(int source, int dest,
			      double act, double delta, double[] weights) {
      double change = learningRate * delta * act;

      weights[source] += change;
      
      ui.write("Weight change for " + source + "->" + dest + ": " + change, 3);
  } // method updateWeight
  
  private void updateOutputWeights() {
      int d, s;
      double delta;
      double[] weights;
      
      for (d = 0; d < nOutputs; d++) {
	  delta = outputDeltas[d];
	  if (nHidden > 0) {
  	    weights = hid2OutWeights[d];
            for (s = 0; s < nHidden; s++)
	        updateWeight(s, d, hidden[s], delta, weights);
	  }
          weights = in2OutWeights[d];
	  for (s = 0; s < nInputs + 1; s++)
	      updateWeight(s, d, inputs[s], delta, weights);
      }
  } // method updateOutputWeights()
  
  private void updateHiddenWeights() {
      int d, s;
      double delta;
      double[] weights;
      
      for (d = 0; d < nHidden; d++) {
	  delta = hiddenDeltas[d];
	  weights = in2HidWeights[d];
          for (s = 0; s < nInputs + 1; s++)
	      updateWeight(s, d, inputs[s], delta, weights);
      }
  } // method updateHiddenWeights()
  
  public void setPercept(int[] percept) {
      int i;
      for (i = 0; i < percept.length; i++)
	  inputs[i] = percept[i];
  } // method setPercept(int[])
  
  public void respond(int[] percept, int age) {
      int response = selectAction(percept, age);
//      ui.write("Selected action " + response);
      attempt(response);
      update(percept, response);
  } // method respond(int[], int)

  public void learn(int[] percept, int reinf) {
//      ui.write("Learning with reinf: " + reinf);
      
      if (fresh) {
	  // Don't learn when first created
	  // (lastPercept and lastAction are wrong).
          fresh = false;
      } else {
	  double Q, targetQ, reducedReinf = reinf * REINF_MULT;
	  
	  // Calculate the target Q
	  targetQ = getNewQ(percept, reducedReinf);
	  ui.write("Target Q = " + targetQ, 1);
	  // Then activate the network to find its Q and
	  // do backprop.
	  run(lastPercept, targetQ, lastAction);
      }
  } // method learn(int[], int)

  private int selectBestAction(int[] percept, int age) {
      String s = "";
      
      ui.write("Selecting best action in memory network", 1);
      
      activate(percept);
      for (int i = 0; i < nOutputs; i++)
	  s += " " + outputs[i];

      ui.write("Output activations: " + s, 1);
      
      return selectBestAction(age);
  } // method selectBestAction(int[], age)
  
  private int selectBestAction(int age) {
      int i, bestI = 0;
      double[] scaledQs = new double[nOutputs];
      
      for (i = 0; i < nOutputs; i++)
	  scaledQs[i] = Math.exp(age * exploitationRate * getQ(i));
      
      Utils.scale(scaledQs);
      
      bestI = Utils.chooseByDistribution(scaledQs);
	  
      ui.write("Best action: " + bestI, 1);
      return bestI;
  } // method selectBestAction(int)

  public void update(int[] percept, int action) {
      lastPercept = percept;
      lastAction = action;
  } // method update(int[], int)

  public int selectAction(int[] percept, int age) {
        return (selectBestAction(percept, age));
  } // method selectAction(int[])

  private int selectRandomAction() {
      return (int)(Math.random() * nOutputs);
  } // method selectRandomAction
  
  private void attempt(int out) {
      ui.write("Memory: attempting " + out, 1);
      getEffector(outputs2Actions[out][0]).attempt(outputs2Actions[out][1]);
  } // attempt(int)

  public boolean read(Tokenizer tok) {
      int i, h, o;
      boolean ok = true;
      
      ui.write("Reading weights into species.");
      try {
	  // Network dimensions.
	  tok.nextToken();
	  nInputs = (int)(tok.nval);
	  tok.nextToken();
	  nHidden = (int)(tok.nval);
	  tok.nextToken();
	  nOutputs = (int)(tok.nval);
	  // Create the weights, etc.
	  init();
          // Input to output
          for (o = 0; o < nOutputs; o++)
	      for (i = 0; i < nInputs + 1; i++) {
		  tok.nextToken();
		  setIn2Out(o, i, tok.nval);
	      }

	  if (nHidden > 0) {
	  // Input to hidden
      	  for (h = 0; h < nHidden; h++)
      	      for (i = 0; i < nInputs + 1; i++) {
		  tok.nextToken();
		  setIn2Hid(h, i, tok.nval);
	      }
	    // Hidden to output
	  for (o = 0; o < nOutputs; o++)
      	      for (h = 0; h < nHidden; h++) {
		  tok.nextToken();
		  setHid2Out(o, h, tok.nval);
	      }
	  }
      }
      catch (IOException e) {
	  System.err.println("Memory: read: " + e);
      }

      return ok;
  } // read(Tokenizer)

  private void writeWeight(PrintWriter out, double wt) {
      if (Math.abs(wt) < .001)
	  out.println("0.0");
      else
	  out.println(wt);
  } // writeWeight(PrintWriter, double)

  public void write(PrintWriter out) {
      int i, h, o;
      out.println(Params.MEMORY_KEY);
      out.println(nInputs + " " + nHidden + " " + nOutputs);
      // Input to output
      for (o = 0; o < nOutputs; o++)
	  for (i = 0; i < nInputs + 1; i++)
	      writeWeight(out, getIn2Out(o, i));

      if (nHidden > 0) {

	  // Input to hidden
	  for (h = 0; h < nHidden; h++)
	      for (i = 0; i < nInputs + 1; i++)
		  writeWeight(out, getIn2Hid(h, i));

	  // Hidden to output
	  for (o = 0; o < nOutputs; o++)
	      for (h = 0; h < nHidden; h++)
		  writeWeight(out, getHid2Out(o, h));
      }
  } // write(PrintWriter)
  
} // class Memory
